import time
from mock import patch
from unittest import TestCase

import test_settings
from shardmonster import api, metadata
from shardmonster.tests.base import ShardingTestCase


class TestShardMetadataStore(ShardingTestCase):
    def setUp(self):
        super(TestShardMetadataStore, self).setUp()
        self._cache_length = 0.05
        api.activate_caching(self._cache_length)


    def tearDown(self):
        super(TestShardMetadataStore, self).tearDown()
        # Deactivate caching by setting a 0 timeout
        api.activate_caching(0)


    @patch('shardmonster.metadata.ShardMetadataStore._query_shards_collection')
    def test_cache_single_shard(self, mock_query):
        expected_shard_metadata = {
            'status': metadata.ShardStatus.AT_REST, 'shard_key': 1}
        mock_query.return_value = [expected_shard_metadata.copy()]

        store = metadata.ShardMetadataStore({'realm': 'dummy-realm'})
        actual_shard_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_shard_metadata, actual_shard_metadata)
        self.assertEquals(1, mock_query.call_count)
    
        # Do another query and ensure we have used a cached value
        actual_shard_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_shard_metadata, actual_shard_metadata)
        self.assertEquals(1, mock_query.call_count)

        # Now sleep for longer the cache timeout and ensure we do another call
        time.sleep(self._cache_length * 2)
        actual_shard_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_shard_metadata, actual_shard_metadata)
        self.assertEquals(2, mock_query.call_count)


    @patch('shardmonster.metadata.ShardMetadataStore._query_shards_collection')
    def test_cache_single_shard_in_flux(self, mock_query):
        expected_shard_metadata = {
            'status': metadata.ShardStatus.MIGRATING_SYNC, 'shard_key': 1}
        mock_query.return_value = [expected_shard_metadata.copy()]

        store = metadata.ShardMetadataStore({'realm': 'dummy-realm'})
        actual_shard_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_shard_metadata, actual_shard_metadata)
        self.assertEquals(1, mock_query.call_count)
    
        # Do another query and ensure we have NOT used a cached value
        actual_shard_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_shard_metadata, actual_shard_metadata)
        self.assertEquals(2, mock_query.call_count)


    @patch('shardmonster.metadata.ShardMetadataStore._query_shards_collection')
    def test_cache_all_shards(self, mock_query):
        expected_metadata_1 = {
            'status': metadata.ShardStatus.AT_REST, 'shard_key': 1}
        expected_metadata_2 = {
            'status': metadata.ShardStatus.AT_REST, 'shard_key': 2}
        mock_query.return_value = [
            expected_metadata_1.copy(), expected_metadata_2.copy()]

        store = metadata.ShardMetadataStore({'realm': 'dummy-realm'})
        actual_metadata = store.get_all_shard_metadata()
        self.assertEquals(
            {1: expected_metadata_1, 2: expected_metadata_2},
            actual_metadata)
        self.assertEquals(1, mock_query.call_count)
    
        # Do another query and ensure we have used a cached value
        actual_metadata = store.get_all_shard_metadata()
        self.assertEquals(
            {1: expected_metadata_1, 2: expected_metadata_2},
            actual_metadata)
        self.assertEquals(1, mock_query.call_count)

        # Do a query for a single shard and ensure we get the cache value
        actual_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_metadata_1, actual_metadata)

        # Now sleep for longer the cache timeout and ensure we do another call
        time.sleep(self._cache_length * 2)
        mock_query.return_value = [expected_metadata_1.copy()]
        actual_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_metadata_1, actual_metadata)
        self.assertEquals(2, mock_query.call_count)

        # The previous call will have only refreshed a single shard. Therefore,
        # a global call will result in another query
        actual_metadata = store.get_all_shard_metadata()
        self.assertEquals(
            {1: expected_metadata_1, 2: expected_metadata_2},
            actual_metadata)
        self.assertEquals(3, mock_query.call_count)


    @patch('shardmonster.metadata.ShardMetadataStore._query_shards_collection')
    def test_cache_all_shards_in_flux(self, mock_query):
        expected_metadata_1 = {
            'status': metadata.ShardStatus.MIGRATING_SYNC, 'shard_key': 1}
        expected_metadata_2 = {
            'status': metadata.ShardStatus.AT_REST, 'shard_key': 2}
        mock_query.return_value = [
            expected_metadata_1.copy(), expected_metadata_2.copy()]

        store = metadata.ShardMetadataStore({'realm': 'dummy-realm'})
        actual_metadata = store.get_all_shard_metadata()
        self.assertEquals(
            {1: expected_metadata_1, 2: expected_metadata_2},
            actual_metadata)
        self.assertEquals(1, mock_query.call_count)
        
        # A query of the second shard should hit the cache
        actual_metadata = store.get_single_shard_metadata(2)
        self.assertEquals(expected_metadata_2, actual_metadata)
        self.assertEquals(1, mock_query.call_count)
    
        # Do another query and we should end up refrshing just a single shard
        mock_query.return_value = [expected_metadata_1.copy()]

        actual_metadata = store.get_all_shard_metadata()
        self.assertEquals(
            {1: expected_metadata_1, 2: expected_metadata_2},
            actual_metadata)
        self.assertEquals(2, mock_query.call_count)
        mock_query.assert_called_with(1)

        # Another full query should skip the cache
        actual_metadata = store.get_all_shard_metadata()
        self.assertEquals(
            {1: expected_metadata_1, 2: expected_metadata_2},
            actual_metadata)
        self.assertEquals(3, mock_query.call_count)


    def test_query(self):
        api.create_realm(
            'dummy-realm', 'some_field', 'dummy_collection',
            'cluster-1/%s' % test_settings.CONN1['db_name'])
        api.set_shard_at_rest('dummy-realm', 1, 'dest1/some_db')
        expected_metadata = {
            'shard_key': 1,
            'location': 'dest1/some_db',
            'realm': 'dummy-realm'
        }
        
        def _trim_results(docs):
            return [{
                'shard_key': doc['shard_key'],
                'location': doc['location'],
                'realm': doc['realm']
            } for doc in docs]


        store = metadata.ShardMetadataStore({'name': 'dummy-realm'})

        results = _trim_results(store._query_shards_collection())
        self.assertEquals([expected_metadata], results)

        results = _trim_results(store._query_shards_collection(1))
        self.assertEquals([expected_metadata], results)

        results = _trim_results(store._query_shards_collection(2))
        self.assertEquals([], results)


        store = metadata.ShardMetadataStore({
            'name': 'some-other-realm'})
        results = _trim_results(store._query_shards_collection())
        self.assertEquals([], results)

        results = _trim_results(store._query_shards_collection(1))
        self.assertEquals([], results)


    @patch('shardmonster.metadata.ShardMetadataStore._query_shards_collection')
    def test_default_location(self, mock_query):
        default_dest = 'cluster-1/%s' % test_settings.CONN1['db_name']
        expected_shard_metadata = {
            'status': metadata.ShardStatus.AT_REST,
            'location': default_dest,
            'realm': 'dummy-realm',
        }
        mock_query.return_value = []

        store = metadata.ShardMetadataStore({
            'name': 'dummy-realm',
            'default_dest': default_dest})
        actual_shard_metadata = store.get_single_shard_metadata(1)
        self.assertEquals(expected_shard_metadata, actual_shard_metadata)
        self.assertEquals(1, mock_query.call_count)


    def test_get_location_ordering(self):
        # Exposes a bug that was found in caching and default locations
        api.create_realm(
            'dummy-realm', 'some_field', 'dummy_collection',
            'cluster-1/some_db')
        api.set_shard_at_rest('dummy-realm', 1, 'dest2/some_db')
        realm = metadata._get_realm_for_collection('dummy_collection')
        meta = metadata._get_metadata_for_shard(realm, 2)
        expected_meta = {
            'status': metadata.ShardStatus.AT_REST,
            'realm': 'dummy-realm',
            'location': 'cluster-1/some_db'
        }
        self.assertEquals(meta, expected_meta)

        all_locations = metadata._get_all_locations_for_realm(realm)
        self.assertEquals([], all_locations['cluster-1/some_db'].contains)
        self.assertEquals([], all_locations['cluster-1/some_db'].excludes)
        self.assertEquals([1], all_locations['dest2/some_db'].contains)
        self.assertEquals([], all_locations['dest2/some_db'].excludes)


class TestGetRealm(ShardingTestCase):
    def setUp(self):
        super(TestGetRealm, self).setUp()
        self._cache_length = 0.05
        api.activate_caching(self._cache_length)


    def tearDown(self):
        super(TestGetRealm, self).tearDown()
        # Deactivate caching by setting a 0 timeout
        api.activate_caching(0)


    @patch('shardmonster.metadata._get_realm_coll')
    def test_caching(self, mock_get_realm_coll):
        shard_data = {'shard_field': 'domain'}
        mock_get_realm_coll.return_value.find.return_value = [shard_data]

        result = metadata._get_realm_for_collection('bob')
        self.assertEquals(shard_data, result)
        self.assertEquals(1, mock_get_realm_coll.call_count)

        result = metadata._get_realm_for_collection('bob')
        self.assertEquals(shard_data, result)
        self.assertEquals(1, mock_get_realm_coll.call_count)
        
        time.sleep(self._cache_length * 2)
        result = metadata._get_realm_for_collection('bob')
        self.assertEquals(shard_data, result)
        self.assertEquals(2, mock_get_realm_coll.call_count)


class TestLocationMetadata(TestCase):
    def test_repr(self):
        meta = metadata.LocationMetadata("somewhere/banana")
        self.assertEquals(
            "LocationMetadata(somewhere/banana, contains: [], excludes: [])",
            repr(meta))
        meta.contains += [1, 2, 3, 4, 5, 6]
        meta.excludes.append(9)
        self.assertEquals(
            "LocationMetadata(somewhere/banana, "\
            "contains: [1,2,3,4,5...], excludes: [9])",
            repr(meta))



"""Contains everything to do with making Mongo operations work across multiple
clusters.
"""
import bson
import numbers
import time

from shardmonster.connection import get_connection, parse_location
from shardmonster.metadata import (
    _get_shards_coll, ShardStatus, _get_realm_for_collection,
    _get_location_for_shard, _get_all_locations_for_realm,
    _get_metadata_for_shard)


def _create_collection_iterator(collection_name, query, with_options={}):
    """Creates an iterator that returns collections and queries that can then
    be used to perform multishard operations:

        for collection, query in _create_collection_iterator(...):
            for doc in collection.find(query):
                yield doc

    This does all the hardwork of figuring out what collections to query and how
    to adjust the query to account for any shards that are currently moving.
    """
    realm = _get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']

    shard_key = _get_query_target(collection_name, query)
    if shard_key:
        location = _get_location_for_shard(realm, shard_key)
        locations = {location.location: location}
    else:
        locations = _get_all_locations_for_realm(realm)

    for location, location_meta in locations.iteritems():
        cluster_name, database_name = parse_location(location)
        connection = get_connection(cluster_name)
        collection = connection[database_name][collection_name]
        if with_options:
            collection = collection.with_options(**with_options)
        if location_meta.excludes:
            if len(location_meta.excludes) == 1:
                query = {'$and': [
                    query, {shard_field: {'$ne': location_meta.excludes[0]}}]}
            else:
                raise Exception('Multiple shards in transit. Aborting')
        yield collection, query
        if location_meta.excludes:
            query = query['$and'][0]


class MultishardCursor(object):
    def __init__(
            self, collection_name, query, *args, **kwargs):
        self.query = query
        self.collection_name = collection_name
        self.args = args
        self.kwargs = kwargs
        self._hint = kwargs.pop('_hint', None)
        self.with_options = kwargs.pop('with_options', {})
        self._prepared = False
        self._skip = 0


    def _create_collection_iterator(self):
        return _create_collection_iterator(
            self.collection_name, self.query, self.with_options)


    def _prepare_for_iteration(self):
        # The multishard cursor has to keep track of a surprising amount of
        # state. When we want to evaluate a multishard cursor the list of
        # queries that need to be performed (and against which collection) is
        # created. This is then used as a basis of iteration and the fact that
        # cursors are changed during iterations is largely not obvious to the
        # end user of this MultishardCursor.
        self._queries_pending = list(self._create_collection_iterator())
        self._cached_results = None
        self._next_cursor()
        self._prepared = True
        self._skipped = 0


    def _next_cursor(self):
        collection, query = self._queries_pending.pop(0)
        # Skip is implemented by getting results back and then applying the skip.
        # In this situation the limit must be increased before doing the query
        if self._skip and self.kwargs.get('limit'):
            query_kwargs = self.kwargs.copy()
            query_kwargs['limit'] = query_kwargs['limit'] + self._skip
        else:
            query_kwargs = self.kwargs
        cursor = collection.find(query, *self.args, **query_kwargs)
        if self._hint:
            cursor = cursor.hint(self._hint)
        self._current_cursor = cursor


    def __iter__(self):
        return self


    def __len__(self):
        return self.count()


    def next(self):
        res = self._next()
        return res


    def _next(self):
        if not self._prepared:
            self.evaluate()

        if self._skip:
            to_skip = self._skip - self._skipped
        else:
            to_skip = 0

        for _ in range(to_skip):
            self._skipped += 1
            self._next_result()

        return self._next_result()


    def _next_result(self):
        if self._cached_results:
            return self._cached_results.pop(0)

        try:
            return self._current_cursor.next()
        except StopIteration:
            if self._queries_pending:
                self._next_cursor()
                return self.next()
            else:
                raise


    def limit(self, limit):
        self.kwargs['limit'] = limit
        return self


    def skip(self, skip):
        self._skip = skip
        return self


    def sort(self, key_or_list, direction=None):
        if direction:
            self.kwargs['sort'] = [(key_or_list, direction)]
        else:
            self.kwargs['sort'] = key_or_list
        return self


    def clone(self):
        return MultishardCursor(
            self.collection_name, self.query, _hint=self._hint,
            *self.args, **self.kwargs)


    def __getitem__(self, i):
        if isinstance(i, int):
            new_cursor = self.clone()
            new_cursor.limit(1)
            new_cursor.skip(i)
            return list(new_cursor)[0]
        else:
            new_cursor = self.clone()
            new_cursor.skip(i.start or 0)
            if i.stop:
                new_cursor.limit(i.stop - (i.start or 0))
            else:
                new_cursor.limit(0)

            return new_cursor


    def evaluate(self):
        self._prepare_for_iteration()
        if len(self._queries_pending) == 0:
            # By the time this code is reached the first query will have been
            # popped off to create the first cursor. Therefore, 0 pending
            # queries implies this is being run against a single server.
            # Running against a single server means we can rely on the
            # server to do sorting/limiting for us. Otherwise, we have to get
            # all results in memory and apply limits etc here.
            return

        if 'sort' in self.kwargs:
            # Note: This is quite inefficient. In an ideal world this would pass
            # the sort through to each cluster and do the sort at that end and
            # then do a merge sort to save on memory. However, that is more
            # complex and I'd rather this was 100% correct and bloated
            # in memory.
            all_results = list(self)
            def comparator(d1, d2):
                for key, sort_order in self.kwargs['sort']:
                    if d1[key] < d2[key]:
                        return -sort_order
                    elif d1[key] > d2[key]:
                        return sort_order
                return 0
                
            self._cached_results = list(sorted(all_results, cmp=comparator))

        if self.kwargs.get('limit'):
            # Note: This is also inefficient. This gets back all the results and
            # then applies the limit. Again, correctness over efficiency.
            self._cached_results = list(self)[:self.kwargs['limit']]


    def count(self, **count_kwargs):
        total = 0
        for collection, query in self._create_collection_iterator():
            cursor = collection.find(query, *self.args, **self.kwargs)
            if self._hint:
                cursor = cursor.hint(self._hint)
            total += cursor.count(**count_kwargs)
        if self.kwargs.get('limit'):
            return min(self.kwargs['limit'], total)
        else:
            return total


    def rewind(self):
        self._cached_results = None
        self._current_cursor = None
        self._queries_pending = None
        self._prepared = False


    def hint(self, index):
        self._hint = index
        return self


    @property
    def alive(self):
        # Alive has to check the current cursor that is being used - if the
        # current user is not alive then there is a chance that the next cursor
        # could be alive and so we must move onto the next cursor and do
        # the check again.
        if not self._prepared:
            self.evaluate()
        current_alive = self._current_cursor.alive or self._cached_results
        if not current_alive and self._queries_pending:
            self._next_cursor()
            return self.alive

        return current_alive


def _create_multishard_iterator(collection_name, query, *args, **kwargs):
    return MultishardCursor(collection_name, query, *args, **kwargs)


def multishard_find(collection_name, query, *args, **kwargs):
    return _create_multishard_iterator(collection_name, query, *args, **kwargs)


def multishard_find_one(collection_name, query, **kwargs):
    kwargs['limit'] = 1
    cursor = _create_multishard_iterator(collection_name, query, **kwargs)
    try:
        return cursor.next()
    except StopIteration:
        return None


def multishard_insert(
        collection_name, doc_or_docs, with_options={}, *args, **kwargs):
    # TODO Remove this and use insert_one/insert_many to comply with new
    # pymongo deprecations
    is_multi_insert = isinstance(doc_or_docs, list)
    if not is_multi_insert:
        all_docs = [doc_or_docs]
    else:
        all_docs = doc_or_docs

    _wait_for_pause_to_end(collection_name, doc_or_docs)
    realm = _get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']
    for doc in all_docs:
        if shard_field not in doc:
            raise Exception(
                'Cannot insert document without shard field (%s) present'
                % shard_field)

    # Inserts can use our generic collection iterator with a specific query
    # that is guaranteed to return exactly one collection.
    # TODO This makes a multi-insert into lots of small inserts. This could be
    # optimised. For now, we'll see if this is OK.
    result = []
    for doc in all_docs:
        simple_query = {shard_field: doc[shard_field]}
        (collection, _), = _create_collection_iterator(
            collection_name, simple_query, with_options)
        result.append(collection.insert(doc, *args, **kwargs))
    if not is_multi_insert:
        return result[0]
    return result


def _is_valid_type_for_sharding(value):
    return isinstance(value, (numbers.Integral, basestring, bson.ObjectId))


def _get_query_target(collection_name, query):
    """Gets out the targetted shard key from the query if there is one.
    Otherwise, returns None.
    """
    realm = _get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']

    if shard_field in query and _is_valid_type_for_sharding(query[shard_field]):
        return query[shard_field]
    return None


def _should_pause_write(collection_name, query):
    realm = _get_realm_for_collection(collection_name)

    shard_key = _get_query_target(collection_name, query)
    if shard_key:
        meta = _get_metadata_for_shard(realm, shard_key)
        return \
            meta['status'] == ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION
    else:
        paused_query = {
            'realm': realm['name'],
            'status': ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION
        }
        shards_coll = _get_shards_coll()
        return shards_coll.find(paused_query).count() > 0


def _wait_for_pause_to_end(collection_name, query):
    while _should_pause_write(collection_name, query):
        time.sleep(0.05)


def _get_collection_for_targetted_upsert(
        collection_name, query, update, with_options={}):
    shard_key = _get_query_target(collection_name, update['$set'])
    realm = _get_realm_for_collection(collection_name)
    location = _get_location_for_shard(realm, shard_key)

    cluster_name, database_name = parse_location(location.location)
    connection = get_connection(cluster_name)
    collection = connection[database_name][collection_name]
    if with_options:
        collection = collection.with_options(with_options)
    return collection


def multishard_update(collection_name, query, update, with_options={}, **kwargs):
    _wait_for_pause_to_end(collection_name, query)
    overall_result = None
    # If this is an upsert then we check the update to see if it might contain
    # the shard key and use that for the collection iterator. Otherwise,
    # we can end up doing an upsert against all clusters... which results in lots
    # of documents all over the place.
    if (kwargs.get('upsert', False) and '$set' in update and
        _get_query_target(collection_name, update['$set'])):
        # Can't use the normal collection iteration method as it would use the
        # wrong query. Instead, get a specific collection and turn it into the
        # right format.
        collection = _get_collection_for_targetted_upsert(
            collection_name, query, update, with_options)
        collection_iterator = [(collection, query)]
    else:
        collection_iterator = _create_collection_iterator(
            collection_name, query, with_options)

    for collection, targetted_query in collection_iterator:
        result = collection.update(targetted_query, update, **kwargs)
        if not overall_result:
            overall_result = result
        else:
            overall_result['n'] += result['n']

    return overall_result


def multishard_remove(collection_name, query, with_options={}, **kwargs):
    _wait_for_pause_to_end(collection_name, query)
    overall_result = None
    collection_iterator = _create_collection_iterator(
        collection_name, query, with_options)
    for collection, targetted_query in collection_iterator:
        result = collection.remove(targetted_query, **kwargs)
        if not overall_result:
            overall_result = result
        else:
            overall_result['n'] += result['n']

    return overall_result


def multishard_aggregate(
        collection_name, pipeline, with_options={}, *args, **kwargs):
    realm = _get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']
    if '$match' not in pipeline[0]:
        raise Exception(
            'Sharded aggregation needs match in the first part of the pipeline')
    if shard_field not in pipeline[0]['$match']:
        raise Exception(
            'Cannot perform aggregation without shard field (%s) present'
            % shard_field)

    # To avoid aggregation needing to be recreated in this client we limit
    # aggregation to only one cluster.
    match_query = pipeline[0]['$match']
    (collection, _), = _create_collection_iterator(
        collection_name, match_query, with_options)

    # TODO: useCursor needs to be False until support for Mongo2.4 is removed
    return collection.aggregate(pipeline, useCursor=False, *args, **kwargs)


def multishard_save(collection_name, doc, with_options={}, *args, **kwargs):
    _wait_for_pause_to_end(collection_name, doc)
    realm = _get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']
    if shard_field not in doc:
        raise Exception(
            'Cannot save document without shard field (%s) present'
            % shard_field)

    # Inserts can use our generic collection iterator with a specific query
    # that is guaranteed to return exactly one collection.
    simple_query = {shard_field: doc[shard_field]}
    (collection, _), = _create_collection_iterator(
        collection_name, simple_query, with_options)

    return collection.save(doc, *args, **kwargs)


def multishard_ensure_index(collection_name, *args, **kwargs):
    collection_iterator = _create_collection_iterator(collection_name, {})

    for collection, _ in collection_iterator:
        collection.ensure_index(*args, **kwargs)



import logging
import pymongo
import threading
import time

logger = logging.getLogger("shardmonster")

CLUSTER_CACHE_LENGTH = 10 * 60 # Cache URI lookups for 10 minutes

_connection_cache = {}
_cluster_uri_cache = {}
_controlling_db = None


def _connect_to_mongo(uri):
    return pymongo.MongoClient(uri)


def connect_to_controller(uri, db_name):
    """Connects to the controlling database. This contains information about
    the realms, shards and clusters.

    :param str uri: The Mongo URI to connect to. This should typically detail
        several replica members to ensure connectivity.
    :param str db_name: The name of the database to connect to on the given
        replica set.
    """
    global _controlling_db
    _controlling_db = _connect_to_mongo(uri)[db_name]


def _make_connection(cluster_name):
    uri = get_cluster_uri(cluster_name)
    return _connect_to_mongo(uri)


def get_controlling_db():
    """Gets a reference to the database that is the controller for sharding.
    """
    global _controlling_db
    if not _controlling_db:
        raise Exception(
            'Call connect_to_controller before attempting to get a connection')
    return _controlling_db


def get_connection(cluster_name):
    global _connection_cache
    key = '%s:%s' % (threading.current_thread(), cluster_name)
    if key not in _connection_cache:
        connection = _make_connection(cluster_name)
        _connection_cache[key] = connection
        return connection

    connection = _connection_cache[key]
    return connection


def _get_cluster_coll():
    db = get_controlling_db()
    return db.clusters


def add_cluster(name, uri):
    """Adds a cluster with a specific name to the clusters shardmonster is aware
    of.
    """
    coll = _get_cluster_coll()
    coll.insert({'name': name, 'uri': uri})


def ensure_cluster_exists(name, uri):
    """Ensures that a cluster with the given name exists. If it doesn't exist,
    a new cluster definition will be created using name and uri. If it does
    exist then no changes will be made.
    
    :param str name: The name of the cluster
    :param str uri: The URI to use for the cluster
    """
    coll = _get_cluster_coll()
    cursor = coll.find({'name': name})
    if not cursor.count():
        add_cluster(name, uri)
    else:
        existing = cursor[0]
        if existing['uri'] != uri:
            logger.warn(
                "Cluster in database does not match cluster being configured. "
                "This is normally OK if clusters are being moved about."
            )

def get_cluster_uri(name):
    """Gets the URI of the cluster with the given name.

    Caches all lookups for ~10 minutes.
    """
    global _cluster_uri_cache
    now = time.time()
    if name not in _cluster_uri_cache or _cluster_uri_cache[name][1] <= now:
        coll = _get_cluster_coll()
        cluster = coll.find_one({'name': name})
        if not cluster:
            raise Exception('Cluster %s has not been configured' % name)
        uri = cluster['uri']
        expiry = now + CLUSTER_CACHE_LENGTH
        _cluster_uri_cache[name] = (uri, expiry)

    return _cluster_uri_cache[name][0]


def parse_location(location):
    """Parses a location of the form cluster/database into the two parts and
    returns them as a tuple

       >>> parse_location("cluster1/some_db")
       ("cluster1", "some_db")
    """
    if location.count('/') != 1:
        raise Exception('Invalid location %s' % location)
    cluster, db = location.split('/')
    return cluster, db



from shardmonster.connection import (
    get_cluster_uri, _get_cluster_coll, ensure_cluster_exists)
from shardmonster.tests.base import ShardingTestCase


class TestCluster(ShardingTestCase):
    def test_ensure_cluster_exists(self):
        # Trying to get a none-existent realm should blow up
        with self.assertRaises(Exception) as catcher:
            uri = get_cluster_uri('best-cluster')
        self.assertEquals(
            catcher.exception.message,
            'Cluster best-cluster has not been configured')

        ensure_cluster_exists(
            'best-cluster', 'mongodb://localhost:27017')
        uri = get_cluster_uri('best-cluster')
        self.assertEquals('mongodb://localhost:27017', uri)

        # Try creating the cluster again, ensure it doesn't blow up or create a
        # duplicate
        ensure_cluster_exists(
            'best-cluster', 'mongodb://localhost:27017')
        uri = get_cluster_uri('best-cluster')
        self.assertEquals('mongodb://localhost:27017', uri)

        coll = _get_cluster_coll()
        # Two clusters exist due to the base class
        self.assertEquals(3, coll.count())

    # TODO Changing clusters



"""
This is an integration test suite that attempts to move data between two shards
a few times whilst also doing inserts and updates. It then verifies that the
data makes sense at the very end.

This works by creating documents of the form:

    {
        account_id: 1,
        some_counter: 1,
        some_key: 1,
    }

It then performs sharding on account ID whilst doing changes to some_counter.
"""
import sys
import time


import test_settings
from shardmonster import api, sharder
from shardmonster.tests.base import ShardingTestCase


class TestWholeThing(ShardingTestCase):
    def setUp(self):
        super(TestWholeThing, self).setUp()
        self._create_indices()
        self._make_collections_shard_aware()
        api.activate_caching(0.5)


    def tearDown(self):
        super(TestWholeThing, self).tearDown()
        # Deactivate caching by setting a 0 timeout
        api.activate_caching(0)

 
    def _make_collections_shard_aware(self):
        self.unwrapped_dummy_1 = self.db1.dummy
        self.unwrapped_dummy_2 = self.db2.dummy
        self.db1.dummy = api.make_collection_shard_aware('dummy')
        self.db2.dummy = api.make_collection_shard_aware('dummy')
        self.sharded_coll = self.db1.dummy


    def _create_indices(self):
        self.db1.dummy.ensure_index(
            [('account_id', 1), ('some_key', 1)],
            unique=True)
        self.db2.dummy.ensure_index(
            [('account_id', 1), ('some_key', 1)],
            unique=True)


    def _prepare_account_data(self, db, account_id, key_range):
        records = []
        for i in key_range:
            record = {
                'account_id': account_id,
                'some_key': i,
                'counter': 1
            }
            record['_id'] = db.dummy.insert(record)
            records.append(record)
        return records


    def _prepare_realms(self):
        api.create_realm('dummy', 'account_id', 'dummy', 'dest1/test_sharding')
    
    def _modify_data(
            self, account_1, account_2, start_insert_id, number_inserts):
        # Increment all the counters a few times
        for cnt in range(3):
            for record in account_1:
                self.sharded_coll.update(
                    {'account_id': 1, 'some_key': record['some_key']},
                    {'$inc': {'counter': 1}},
                )
                record['counter'] += 1

        # Delete all the counters that are divisible by 97 (including 0)
        to_delete = []
        for record in account_1:
            if record['some_key'] % 97 == 0:
                to_delete.append(record)
        for record in to_delete:
            account_1.remove(record)
            self.sharded_coll.remove(
                {'account_id': 1, 'some_key': record['some_key']})

        # Add some additional records
        account_1 += self._prepare_account_data(
            self.db1, 1,
            xrange(start_insert_id, start_insert_id + number_inserts))
        account_2 += self._prepare_account_data(
            self.db1, 2,
            xrange(start_insert_id, start_insert_id + number_inserts))

    def _verify_end_state(self, account_1, account_2, original_1, original_2):
        # Fetch the data from the second server and check all the records match
        account_1_actual = list(original_2.find({'account_id': 1}))
        account_1_actual = list(sorted(
            account_1_actual, key=lambda r: r['some_key']))

        if account_1 != account_1_actual:
            print 'Account 1 Expected'
            print '------------------'
            for doc in account_1:
                print '(%s, %s) -> %d' % (
                    doc['account_id'], doc['some_key'], doc['counter'])
            print 'Account 1 Actual'
            print '----------------'
            for doc in account_1_actual:
                print '(%s, %s) -> %d    (%s)' % (
                    doc['account_id'], doc['some_key'], doc['counter'], doc['_id'])
        self.assertEquals(account_1, account_1_actual)

        # There should be no data for the 1st account at the source
        self.assertEquals(0, original_1.find({'account_id': 1}).count())

    def _attempt_migration(self, num_records):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.set_shard_at_rest('dummy', 2, "dest1/test_sharding")

        account_1 = self._prepare_account_data(
            self.db1, 1, xrange(0, num_records))
        account_2 = self._prepare_account_data(
            self.db1, 2, xrange(0, num_records))

        shard_manager = sharder._begin_migration(
            'dummy', 1, "dest2/test_sharding")
        self._modify_data(account_1, account_2, num_records, num_records)
        while not shard_manager.is_finished():
            time.sleep(0.01)
        self._verify_end_state(
            account_1, account_2, self.unwrapped_dummy_1, self.unwrapped_dummy_2)

        # Check that the data for the other account has remained intact and in
        # the same place
        account_2_actual = list(self.unwrapped_dummy_1.find({'account_id': 2}))
        account_2_actual = list(sorted(
            account_2_actual, key=lambda r: r['some_key']))

        self.assertEquals(account_2, account_2_actual)

        # Now migrate back to the source
        print 'Now migrate backwards...'
        shard_manager = sharder._begin_migration(
            'dummy', 1, "dest1/test_sharding")
        self._modify_data(account_1, account_2, num_records * 2, num_records)
        while not shard_manager.is_finished():
            time.sleep(0.01)
        self._verify_end_state(
            account_1, account_2, self.unwrapped_dummy_2, self.unwrapped_dummy_1)


# Make the tests dynamically so that test runners break it up :)
for i, records in enumerate(test_settings.INTEGRATION_TEST_RUNS):
    def _make_test(run, run_records):
        def test_method(self):
            start_time = time.time()
            msg = '[%d/%d] Testing sharding with %d initial documents\n' % (
                run, len(test_settings.INTEGRATION_TEST_RUNS), run_records
            )
            sys.stderr.write(msg)

            self._attempt_migration(run_records)

            end_time = time.time()
            sys.stderr.write(
                '  Took %.2fs\n' % (end_time - start_time))
    
        name = 'test_%05d_%05d' % (run_records, i)
        test_method.__doc__ = 'Migration of %d docs, run #%d' % (run_records, i)
        return name, test_method
    name, tester = _make_test(i + 1, records)
    setattr(TestWholeThing, name, tester)

# Set tester to None at the end so that nose doesn't pick this up as a test to
# run
tester = None



import bson
from mock import patch
from pymongo.errors import OperationFailure

from shardmonster import api, operations
from shardmonster.tests.base import ShardingTestCase

class TestStandardMultishardOperations(ShardingTestCase):
    def setUp(self):
        super(TestStandardMultishardOperations, self).setUp()
        api.set_shard_at_rest('dummy', 1, 'dest1/test_sharding')
        api.set_shard_at_rest('dummy', 2, 'dest2/test_sharding')

    def test_multishard_find(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        c = operations.multishard_find('dummy', {'y': 1})
        results = sorted(list(c), key=lambda d: d['x'])
        self.assertEquals([doc1, doc2], results)

    def test_multishard_find_args(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        c = operations.multishard_find('dummy', {'y': 1}, {'x': 1, '_id': 0})
        results = sorted(list(c), key=lambda d: d['x'])
        self.assertEquals([{'x': 1}, {'x': 2}], results)

    def test_multishard_find_with_sort(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', 1)])
        self.assertEquals([doc1, doc2, doc3, doc4], list(results))

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', -1), ('y', 1)])
        self.assertEquals([doc3, doc4, doc1, doc2], list(results))

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', -1)])
        self.assertEquals([doc2, doc1, doc4, doc3], list(results))

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', -1), ('y', -1)])
        self.assertEquals([doc4, doc3, doc2, doc1], list(results))
    
        # Insert a document the same as doc4 to ensure sorts will cope with
        # things that are basically the same
        doc5 = {'x': 2, 'y': 2, 'z': 1}
        self.db2.dummy.insert(doc5)
        results = operations.multishard_find(
            'dummy', {}, sort=[('x', -1), ('y', -1)])
        results = results[:2]
        self.assertTrue(doc4 in results)
        self.assertTrue(doc5 in results)
    
    def test_multishard_find_with_sort_fn(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)

        results = operations.multishard_find(
            'dummy', {}).sort([('x', 1), ('y', 1)])
        self.assertEquals([doc1, doc2, doc3, doc4], list(results))

        results = operations.multishard_find(
            'dummy', {}).sort([('x', -1), ('y', 1)])
        self.assertEquals([doc3, doc4, doc1, doc2], list(results))

    def test_multishard_find_with_sort_and_limit(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', 1)], limit=3)
        self.assertEquals([doc1, doc2, doc3], list(results))

    def test_multishard_find_one(self):
        r = operations.multishard_find_one('dummy', {'x': 1})
        self.assertEquals(None, r)

        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        r = operations.multishard_find_one('dummy', {'x': 1})
        self.assertEquals(r, doc1)

        r = operations.multishard_find_one('dummy', {'x': 2})
        self.assertEquals(r, doc2)

    def test_multishard_find_with_limit(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)

        results = operations.multishard_find('dummy', {}, limit=3)
        self.assertEquals(3, len(list(results)))

    def test_multishard_find_with_limit_as_method(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)

        results = operations.multishard_find('dummy', {}).limit(3)
        self.assertEquals(3, len(list(results)))

    def test_multishard_find_with_shardkey_present(self):
        # Create an unlikely scenario where data is present in both locations
        # even though the data is "at rest". Then do a query with the shard
        # key present. As the shard key is present the query should be
        # targetted enough to ignore the dummy data
        doc1 = {'x': 1, 'y': 1}
        doc2_bad = {'x': 2, 'y': 1, 'bad': True}
        doc2_good = {'x': 2, 'y': 1, 'bad': False}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2_bad)
        self.db2.dummy.insert(doc2_good)

        results = operations.multishard_find('dummy', {'x': 2, 'y': 1})
        self.assertEquals([doc2_good], list(results))

    def test_insert(self):
        # Perform an insert with shards set to specific locations.
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        operations.multishard_insert('dummy', doc1)
        operations.multishard_insert('dummy', doc2)

        results = list(self.db1.dummy.find({'y': 1}))
        self.assertEquals([doc1], results)

        results = list(self.db2.dummy.find({'y': 1}))
        self.assertEquals([doc2], results)


    def test_insert_list(self):
        # Perform inserts with multiple documents at once
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        operations.multishard_insert('dummy', [doc1, doc2])

        results = list(self.db1.dummy.find({'y': 1}))
        self.assertEquals([doc1], results)

        results = list(self.db2.dummy.find({'y': 1}))
        self.assertEquals([doc2], results)

    def test_insert_with_longs(self):
        # Perform an insert using longs. This tests a specific bug we found
        # during extended testing
        #api.set_shard_at_rest('dummy', 1L, "dest1/test_sharding")
        #api.set_shard_at_rest('dummy', 2L, "dest2/test_sharding")
        doc1 = {'x': 1L, 'y': 1L}
        doc2 = {'x': 2L, 'y': 1L}
        operations.multishard_insert('dummy', doc1)
        operations.multishard_insert('dummy', doc2)

        results = list(self.db1.dummy.find({'y': 1L}))
        self.assertEquals([doc1], results)

        results = list(self.db2.dummy.find({'y': 1L}))
        self.assertEquals([doc2], results)

    def test_multi_update(self):
        # Test that an update will hit multiple clusters at once
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        result = operations.multishard_update('dummy', {}, {'$inc': {'y': 1}})
        self.assertEquals(2, result['n'])

        result, = operations.multishard_find('dummy', {'x': 1})
        self.assertEquals(2, result['y'])

        result, = operations.multishard_find('dummy', {'x': 2})
        self.assertEquals(2, result['y'])

    def test_remove(self):
        # Perform removes with shards set to specific locations.
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        operations.multishard_remove('dummy', {'x': 1, 'y': 1})

        self.assertEquals(0, self.db1.dummy.find({'x': 1}).count())
        self.assertEquals(1, self.db2.dummy.find({'x': 1}).count())

    def test_multi_remove(self):
        # Tests remove across multiple clusters
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        operations.multishard_remove('dummy', {'y': 1})

        self.assertEquals(0, self.db1.dummy.find({}).count())
        self.assertEquals(0, self.db2.dummy.find({}).count())

    def test_aggregate(self):
        for y in range(10):
            doc1 = {'x': 1, 'y': y}
            doc2 = {'x': 2, 'y': y}
            self.db1.dummy.insert(doc1)
            self.db2.dummy.insert(doc2)

        pipeline = [
            {'$match': {'x': 2}},
            {'$group': {'_id': 'total', 's': {'$sum': '$y'}}},
        ]
        result = list(operations.multishard_aggregate('dummy', pipeline))
        self.assertEquals([{'_id': 'total', 's': 45}], result)

    def test_multishard_rewind(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        cursor = operations.multishard_find('dummy', {'y': 1}, sort=[('x', 1)])
        found = cursor.next()
        self.assertEquals((1, 1), (found['x'], found['y']))
        cursor.rewind()
        found = cursor.next()
        self.assertEquals((1, 1), (found['x'], found['y']))

    def test_save(self):
        # A save differs from an insert in that it may perform an insert or it
        # may perform an update, depending on whether an _id exists
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        operations.multishard_save('dummy', doc1)
        operations.multishard_save('dummy', doc2)

        doc1['z'] = 10
        doc2['z'] = 20
        operations.multishard_save('dummy', doc1)
        operations.multishard_save('dummy', doc2)
    
        results = list(self.db1.dummy.find({'y': 1}))
        self.assertEquals([doc1], results)

        results = list(self.db2.dummy.find({'y': 1}))

        self.assertEquals([doc2], results)

    def test_targetted_upsert(self):
        # Tests a bug around targetted upserts going out to all clusters and
        # causing data to appear in multiple places.
        doc1 = {'_id': 'alpha', 'x': 1, 'y': 1}
        operations.multishard_update(
            'dummy', {'_id': 'alpha'}, {'$set': {'x': 1, 'y': 1}}, upsert=True)

        results = list(self.db1.dummy.find({'y': 1}))
        self.assertEquals([doc1], results)

        results = list(self.db2.dummy.find({'y': 1}))
        self.assertEquals([], results)

    def test_hint(self):
        # The easier way to test if a hint is being applied is to apply a bad
        # hint and watch things blow up
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        # This should explode due to a lack of index
        try:
            c = operations.multishard_find('dummy', {'x': 1})
            c = c.hint([('apples', 1)])
            list(c)
        except OperationFailure as e:
            self.assertTrue("bad hint" in str(e))

    def test_indexed_read(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        cursor = operations.multishard_find(
            'dummy', {'y': 1}, sort=[('x', 1), ('y', 1)])
        self.assertEquals(doc1, cursor[0])
        cursor = operations.multishard_find(
            'dummy', {'y': 1}, sort=[('x', -1), ('y', 1)])
        self.assertEquals(doc2, cursor[0])

    def test_unbound_slice(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        c = operations.multishard_find('dummy', {'y': 1})[:]
        results = sorted(list(c), key=lambda d: d['x'])
        self.assertEquals([doc1, doc2], results)

    def test_multishard_find_with_sort_as_single_arg(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        results = operations.multishard_find(
            'dummy', {}).sort('x', 1)
        self.assertEquals([doc1, doc2], list(results))

        results = operations.multishard_find(
            'dummy', {}).sort('x', -1)
        self.assertEquals([doc2, doc1], list(results))

    def test_alive_across_shards(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        c = operations.multishard_find('dummy', {})
        self.assertTrue(c.alive)
        c.next()
        self.assertTrue(c.alive)
        c.next()
        self.assertFalse(c.alive)

    def test_alive_with_sort(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        c = operations.multishard_find('dummy', {}).sort('x', 1)
        self.assertTrue(c.alive)
        c.next()
        self.assertTrue(c.alive)
        c.next()
        self.assertFalse(c.alive)

    def test_multishard_skip(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', 1)]).skip(1)
        self.assertEquals([doc2, doc3, doc4], list(results))

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', 1)]).skip(2)
        self.assertEquals([doc3, doc4], list(results))

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', 1)]).skip(3)
        self.assertEquals([doc4], list(results))

        results = operations.multishard_find(
            'dummy', {}, sort=[('x', 1), ('y', 1)]).skip(4)
        self.assertEquals([], list(results))

    def test_skip_slice(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        c = operations.multishard_find('dummy', {'y': 1})[1:]
        results = sorted(list(c), key=lambda d: d['x'])
        self.assertEquals([doc2], results)

    def test_non_zero_indexing(self):
        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 2, 'y': 1}
        self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc2)

        result = operations.multishard_find('dummy', {'y': 1})[1]
        self.assertEquals(doc2, result)

    def test_skip_beyond_limit(self):
        self.db1.dummy.insert({'x': 1, 'y': 1})
        self.db1.dummy.insert({'x': 1, 'y': 1})
        self.db1.dummy.insert({'x': 1, 'y': 1})
        self.db1.dummy.insert({'x': 1, 'y': 1})
        expected_doc = {'x': 2, 'y': 1}
        self.db2.dummy.insert(expected_doc)

        print expected_doc
        result = operations.multishard_find('dummy', {'y': 1}).limit(1).skip(4)
        self.assertEquals([expected_doc], list(result))


class TestOtherOperations(ShardingTestCase):
    def test_multishard_find_during_migration(self):
        # Indiciate a migration has started on shard #2 and insert a document
        # with the same ID into both databases with slightly different data in
        # each location. Then ensure we only get the result from shard #1
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.set_shard_at_rest('dummy', 2, "dest1/test_sharding")
        # We require a shard at rest on the target server to ensure the location
        # is picked up as a queryable location
        api.set_shard_at_rest('dummy', 3, "dest2/test_sharding")
        api.start_migration('dummy', 2, "dest2/test_sharding")
        doc1 = {'x': 1, 'y': 1}
        doc_id = bson.ObjectId()
        doc2_fresh = {'_id': doc_id, 'x': 2, 'y': 1, 'is_fresh': True}
        doc2_stale = {'_id': doc_id, 'x': 2, 'y': 1, 'is_fresh': False}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2_fresh)
        self.db2.dummy.insert(doc2_stale)

        c = operations.multishard_find('dummy', {'y': 1})
        results = sorted(list(c), key=lambda d: d['x'])
        self.assertEquals([doc1, doc2_fresh], results)

    def test_multishard_find_during_post_migration(self):
        # Indiciate a migration has started on shard #2 and insert a document
        # with the same ID into both databases with slightly different data in
        # each location. Then ensure we only get the result from shard #1
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.set_shard_at_rest('dummy', 2, "dest1/test_sharding")
        api.start_migration('dummy', 2, "dest2/test_sharding")
        api.set_shard_to_migration_status(
            'dummy', 2, api.ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION)
        doc1 = {'x': 1, 'y': 1}
        doc_id = bson.ObjectId()
        doc2_fresh = {'_id': doc_id, 'x': 2, 'y': 1, 'is_fresh': True}
        doc2_stale = {'_id': doc_id, 'x': 2, 'y': 1, 'is_fresh': False}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2_fresh)
        self.db2.dummy.insert(doc2_stale)

        c = operations.multishard_find('dummy', {'y': 1})
        results = sorted(list(c), key=lambda d: d['x'])
        self.assertEquals([doc1, doc2_stale], results)

    def test_insert_to_default_location(self):
        doc1 = {'x': 1, 'y': 1}
        operations.multishard_insert('dummy', doc1)

        results = list(self.db1.dummy.find({'y': 1}))
        self.assertEquals([doc1], results)

    def test_update(self):
        # Put the same document in multiple locations (a mid-migration status)
        # then do an update and ensure that only the correct place has been
        # updated.
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        doc1 = {'x': 1, 'y': 1}
        self.db1.dummy.insert(doc1)

        api.start_migration('dummy', 1, 'dest2/test_sharding')
        api.set_shard_to_migration_status(
            'dummy', 1, api.ShardStatus.MIGRATING_COPY)
        self.db2.dummy.insert(doc1)

        result = operations.multishard_update('dummy', {}, {'$inc': {'y': 1}})
        self.assertEquals(1, result['n'])

        # Query the correct shard first and see that the counter has been
        # incremented
        result, = operations.multishard_find('dummy', {'x': 1})
        self.assertEquals(2, result['y'])

        # Now spoof the metadata such that the system thinks the data is on
        # shard2. The counter should still be 1 here.
        api.set_shard_at_rest('dummy', 1, "dest2/test_sharding", force=True)
        result, = operations.multishard_find('dummy', {'x': 1})
        self.assertEquals(1, result['y'])

    @patch('shardmonster.operations._should_pause_write')
    @patch('shardmonster.operations.time.sleep')
    def test_wait_for_pause_to_end(self, mock_sleep, mock_should_pause):
        mock_should_pause.side_effect = [True, True, False]

        operations._wait_for_pause_to_end("collection", {'field': 1})

        mock_should_pause.assert_called_with("collection", {'field': 1})
        self.assertEquals(3, mock_should_pause.call_count)
        self.assertEquals(2, mock_sleep.call_count)

    def test_should_pause_write(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.set_shard_to_migration_status(
            'dummy', 1, api.ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION)
        self.assertTrue(operations._should_pause_write("dummy", {'x': 1}))

    def test_alive(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        doc1 = {'x': 1, 'y': 1}
        self.db1.dummy.insert(doc1)

        c = operations.multishard_find('dummy', {})
        self.assertTrue(c.alive)

    def test_multishard_count_with_motion(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.set_shard_at_rest('dummy', 2, "dest1/test_sharding")

        doc1 = {'x': 1, 'y': 1}
        doc2 = {'x': 1, 'y': 2}
        doc3 = {'x': 2, 'y': 1}
        doc4 = {'x': 2, 'y': 2}
        self.db1.dummy.insert(doc1)
        self.db1.dummy.insert(doc2)
        self.db1.dummy.insert(doc3)
        self.db1.dummy.insert(doc4)

        results = operations.multishard_find('dummy', {}).count()
        self.assertEquals(4, results)

        # Mimic the shard now being in the second location and there being
        # documents left here
        api.start_migration('dummy', 2, "dest2/test_sharding")
        api.set_shard_to_migration_status(
            'dummy', 2, api.ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION)

        self.db2.dummy.insert(doc3)
        self.db2.dummy.insert(doc4)
    
        results = operations.multishard_find('dummy', {}).count()
        self.assertEquals(4, results)



"""
The actual sharding process for moving data around between servers.

Running
-------

python shardmonster --server=localhost:27017 --db=shardmonster

The server/database is the location of the metadata.
"""
import pymongo
import sys
import threading
import time
from datetime import datetime

from shardmonster import api, metadata
from shardmonster.connection import (
    get_connection, get_controlling_db, parse_location)

STATUS_COPYING = 'copying'
STATUS_SYNCING = 'syncing'
STATUS_SYNCED = 'synced'
ALL_STATUSES = [STATUS_COPYING, STATUS_SYNCING, STATUS_SYNCED]
STATUS_MAP = {STATUS: STATUS for STATUS in ALL_STATUSES}

HEADER = '\033[95m'
OKBLUE = '\033[94m'
OKGREEN = '\033[92m'
WARNING = '\033[93m'
FAIL = '\033[91m'
ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'

def blue(s, *args):
    print OKBLUE + (s % args) + ENDC


def _detail_log(s):
    print '    ', datetime.now(), s


def _get_collection_from_location_string(location, collection_name):
    server_addr, database_name = parse_location(location)
    connection = get_connection(server_addr)
    return connection[database_name][collection_name]


def _do_copy(collection_name, shard_key, insert_throttle=None):
    realm = metadata._get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']

    shards_coll = api._get_shards_coll()
    shard_metadata, = shards_coll.find(
        {'realm': realm['name'], 'shard_key': shard_key})
    if shard_metadata['status'] != metadata.ShardStatus.MIGRATING_COPY:
        raise Exception('Shard not in copy state (phase 1)')

    current_location = shard_metadata['location']
    new_location = shard_metadata['new_location']

    current_collection = _get_collection_from_location_string(
        current_location, collection_name)

    new_collection = _get_collection_from_location_string(
        new_location, collection_name)

    query = {shard_field: shard_key}
    cursor = current_collection.find(query, no_cursor_timeout=True)
    inserted = 0
    try:
        for record in cursor:
            new_collection.insert(record, w=0)
            if inserted % 50000 == 0:
                _detail_log('%d records inserted' % inserted)
            if insert_throttle:
                time.sleep(insert_throttle)
            inserted += 1
    finally:
        cursor.close()

    result = new_collection.database.command('getLastError')
    if result['err']:
        raise Exception('Failed to do copy! Mongo error: %s' % result['err'])


def _get_metadata_for_shard(realm_name, shard_key):
    shards_coll = api._get_shards_coll()
    shard_metadata, = shards_coll.find(
        {'realm': realm_name, 'shard_key': shard_key})
    return shard_metadata


def _get_oplog_pos(collection_name, shard_key):
    """Gets the oplog position for the given collection/shard key combination.
    This is necessary as the oplog will be very different on different clusters.
    """
    realm = metadata._get_realm_for_collection(collection_name)
    shard_metadata = _get_metadata_for_shard(realm['name'], shard_key)

    current_location = shard_metadata['location']
    current_collection = _get_collection_from_location_string(
        current_location, collection_name)
    current_conn = current_collection.database.client

    repl_coll = current_conn['local']['oplog.rs']
    most_recent_op = repl_coll.find({}, sort=[('$natural', -1)])[0]
    ts_from = most_recent_op['ts']
    return ts_from


def _sync_from_oplog(collection_name, shard_key, oplog_pos):
    """Syncs the oplog to within a reasonable timeframe of "now".
    """
    realm = metadata._get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']

    shard_metadata = _get_metadata_for_shard(realm['name'], shard_key)

    current_location = shard_metadata['location']
    new_location = shard_metadata['new_location']

    current_collection = _get_collection_from_location_string(
        current_location, collection_name)

    new_collection = _get_collection_from_location_string(
        new_location, collection_name)

    # Get the connection used by the source collection and use that for the
    # oplog tailing
    conn = current_collection.database.client
    repl_coll = conn['local']['oplog.rs']
    cursor = repl_coll.find(
        {'ts': {'$gte': oplog_pos}},
        cursor_type=pymongo.CursorType.TAILABLE,
        oplog_replay=True)
    cursor = cursor.hint([('$natural', 1)])

    shard_query = {shard_field: shard_key}

    current_namespace = "%s.%s" % (
        current_collection.database.name, current_collection.name)

    for r in cursor:
        if r['ns'] != current_namespace:
            continue

        if r['op'] in ['u', 'i']:
            # Check that this doc is part of our query set
            oid = r.get('o2', r['o'])['_id']
            object_query = {'_id': oid}
            object_query.update(shard_query)
            match = bool(
                current_collection.find(object_query).count())
        elif r['op'] == 'd':
            oid = r.get('o2', r['o'])['_id']
            object_query = {'_id': oid}
            object_query.update(shard_query)
            match = bool(
                new_collection.find(object_query).count())
        else:
            # Notification ops can be ignored.
            continue

        if not match:
            continue

        if r['op'] == 'u':
            blue(' - Updating %s with %s' % (oid, r['o']))
            new_collection.update(
                {'_id': oid}, r['o'], w=1)

        elif r['op'] == 'i':
            try:
                new_collection.insert(r['o'], w=1)
            except pymongo.errors.DuplicateKeyError:
                pass
        elif r['op'] == 'd':
            blue(' - Removing %s' % oid)
            new_collection.remove({'_id': oid}, w=1)

        oplog_pos = r['ts']

    return oplog_pos


def _delete_source_data(collection_name, shard_key, delete_throttle=None):
    realm = metadata._get_realm_for_collection(collection_name)
    shard_field = realm['shard_field']

    shards_coll = api._get_shards_coll()
    shard_metadata, = shards_coll.find(
        {'realm': realm['name'], 'shard_key': shard_key})
    if shard_metadata['status'] != metadata.ShardStatus.POST_MIGRATION_DELETE:
        raise Exception('Shard not in delete state')

    current_location = shard_metadata['location']
    current_collection = _get_collection_from_location_string(
        current_location, collection_name)

    cursor = current_collection.find(
            {shard_field: shard_key}, {'_id': 1},
            no_cursor_timeout=True)
    deleted = 0
    try:
        for doc in cursor:
            current_collection.remove({'_id': doc['_id']})
            if delete_throttle:
                time.sleep(delete_throttle)
            if deleted % 10000 == 0:
                _detail_log('%d records deleted' % deleted)
            deleted += 1

    finally:
        cursor.close()


class ShardMovementThread(threading.Thread):
    def __init__(
            self, collection_name, shard_key, new_location,
            delete_throttle=None, insert_throttle=None):
        self.collection_name = collection_name
        self.shard_key = shard_key
        self.new_location = new_location
        self.exception = None
        self.delete_throttle = delete_throttle
        self.insert_throttle = insert_throttle
        super(ShardMovementThread, self).__init__()


    def run(self):
        try:
            blue('* Starting migration')
            api.start_migration(
                self.collection_name, self.shard_key, self.new_location)

            # Copy phase
            blue('* Doing copy')
            oplog_pos = _get_oplog_pos(self.collection_name, self.shard_key)
            _do_copy(self.collection_name, self.shard_key, self.insert_throttle)

            # Sync phase
            blue('* Initial oplog sync')
            start_sync_time = time.time()
            api.set_shard_to_migration_status(
                self.collection_name, self.shard_key, metadata.ShardStatus.MIGRATING_SYNC)
            oplog_pos = _sync_from_oplog(
                self.collection_name, self.shard_key, oplog_pos)

            # Ensure that the sync has taken at least as long as our caching time
            # to ensure that all writes will get paused at approximately the same
            # time.
            while time.time() < start_sync_time + api.get_caching_duration():
                time.sleep(0.05)
                oplog_pos = _sync_from_oplog(
                    self.collection_name, self.shard_key, oplog_pos)

            # Now all the caching of metadata should be stopped for this shard.
            # We can flip to being paused at destination and wait ~100ms for any
            # pending updates/inserts to be performed. If these are taking longer
            # than 100ms then you are in a bad place and should rethink sharding.
            blue('* Pausing at destination')
            api.set_shard_to_migration_status(
                self.collection_name, self.shard_key,
                metadata.ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION)
            time.sleep(0.1)
            
            blue('* Syncing oplog once more')
            _sync_from_oplog(
                self.collection_name, self.shard_key, oplog_pos)

            # Delete phase
            blue('* Doing deletion')
            api.set_shard_to_migration_status(
                self.collection_name, self.shard_key,
                metadata.ShardStatus.POST_MIGRATION_DELETE)
            _delete_source_data(
                self.collection_name, self.shard_key,
                delete_throttle=self.delete_throttle)

            api.set_shard_at_rest(
                self.collection_name, self.shard_key, self.new_location,
                force=True)

            blue('* Done')
        except:
            self.exception = sys.exc_info()
            raise


class ShardMovementManager(object):
    def __init__(
            self, collection_name, shard_key, new_location,
            delete_throttle=None, insert_throttle=None):
        self.collection_name = collection_name
        self.shard_key = shard_key
        self.new_location = new_location
        self.delete_throttle = delete_throttle
        self.insert_throttle = insert_throttle


    def start_migration(self):
        self._migration_thread = ShardMovementThread(
            self.collection_name, self.shard_key, self.new_location,
            delete_throttle=self.delete_throttle,
            insert_throttle=self.insert_throttle)
        self._migration_thread.start()


    def is_finished(self):
        if self._migration_thread.exception:
            raise Exception(
                'Migration failed %s' % (self._migration_thread.exception,))
        return not self._migration_thread.is_alive()


def _begin_migration(
        collection_name, shard_key, new_location, delete_throttle=None,
        insert_throttle=None):
    if metadata.are_migrations_happening():
        raise Exception(
            'Cannot start migration when another migration is in progress')
    manager = ShardMovementManager(
        collection_name, shard_key, new_location,
        delete_throttle=delete_throttle, insert_throttle=insert_throttle)
    manager.start_migration()
    return manager


def do_migration(
        collection_name, shard_key, new_location, delete_throttle=None,
        insert_throttle=None):
    """Migrates the data with the given shard key in the given collection to
    the new location. E.g.

        >>> do_migration('some_collection', 1, 'cluster-1/some_db')

    Would migrate everything from some_collection where the shard field is set
    to 1 to the database some_db on cluster-1.

    :param str collection_name: The name of the collection to migrate
    :param shard_key: The key of the shard that is to be moved
    :param str new_location: Location that the shard should be moved to in the
        format "cluster/database".
    :param float delete_throttle: This is the length of pause that will be
        applied after each deletion of a source document.
    :param float insert_throttle: This is the length of pause that will be
        applied after each insert of a new document to the destination.

    This method blocks until the migration is completed.
    """
    manager = _begin_migration(
        collection_name, shard_key, new_location,
        delete_throttle=delete_throttle, insert_throttle=insert_throttle)
    while not manager.is_finished():
        time.sleep(0.01)



from shardmonster.api import (
    ensure_realm_exists, set_shard_at_rest, start_migration, where_is)
from shardmonster.metadata import _get_realm_for_collection, _get_realm_coll
from shardmonster.tests.base import ShardingTestCase


class TestRealm(ShardingTestCase):
    def test_ensure_realm_exists(self):
        # Trying to get a none-existent realm should blow up
        with self.assertRaises(Exception) as catcher:
            realm = _get_realm_for_collection('some_collection')
        self.assertEquals(
            catcher.exception.message,
            'Realm for collection some_collection does not exist')

        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'cluster-1/db')
        realm = _get_realm_for_collection('some_collection')
        self.assertEquals('some_realm', realm['name'])

        # Try creating the realm again, ensure it doesn't blow up or create a
        # duplicate
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'cluster-1/db')
        realm = _get_realm_for_collection('some_collection')
        self.assertEquals('some_realm', realm['name'])

        coll = _get_realm_coll()
        self.assertEquals(2, coll.count()) # One realm exists due to test base


    def test_ensure_changing_realm_breaks(self):
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'cluster-1/db')

        with self.assertRaises(Exception) as catcher:
            ensure_realm_exists(
                'some_realm', 'some_other_field', 'some_collection',
                'cluster-1/db')
        self.assertEquals(
            catcher.exception.message, 'Cannot change realm')


    def test_one_realm_per_collection(self):
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'cluster-1/db')

        with self.assertRaises(Exception) as catcher:
            ensure_realm_exists(
                'some_other_realm', 'some_other_field', 'some_collection',
                'cluster-1/db')
        self.assertEquals(
            catcher.exception.message,
            'Realm for collection some_collection already exists')


    def test_cannot_move_to_same_location(self):
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'dest1/db')

        set_shard_at_rest('some_realm', 1, 'dest1/db')

        with self.assertRaises(Exception) as catcher:
            start_migration('some_realm', 1, 'dest1/db')
        self.assertEquals(
            catcher.exception.message, 'Shard is already at dest1/db')


    def test_set_shard_at_rest_bad_location(self):
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'cluster-1/db')

        with self.assertRaises(Exception) as catcher:
            set_shard_at_rest('some_realm', 1, 'bad-cluster/db')
        self.assertEquals(
            catcher.exception.message,
            'Cluster bad-cluster has not been configured')


    def test_set_shard_at_rest_when_already_at_rest(self):
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'dest1/db')
        set_shard_at_rest('some_realm', 1, 'dest1/db')

        with self.assertRaises(Exception) as catcher:
            set_shard_at_rest('some_realm', 1, 'dest2/db')
        self.assertEquals(
            catcher.exception.message,
            'Shard with key 1 has already been placed. Use force=true if '
            'you really want to do this')

        # Forcing is should work
        set_shard_at_rest('some_realm', 1, 'dest2/db', force=True)


    def test_where_is(self):
        ensure_realm_exists(
            'some_realm', 'some_field', 'some_collection', 'dest1/db')
        set_shard_at_rest('some_realm', 1, 'dest2/db')

        # Specific location
        self.assertEquals('dest2/db', where_is('some_collection', 1))
        # Default location
        self.assertEquals('dest1/db', where_is('some_collection', 2))



import time

from shardmonster.connection import (
    _cluster_uri_cache, _get_cluster_coll, get_controlling_db)

class ShardStatus(object):
    def __init__(self):
        raise Exception('Enum. Do not instantiate')

    AT_REST = "at-rest"
    MIGRATING_COPY = "migrating-copy"
    MIGRATING_SYNC = "migrating-sync"
    POST_MIGRATION_PAUSED_AT_SOURCE = "post-migration-paused-at-source"
    POST_MIGRATION_PAUSED_AT_DESTINATION = "post-migration-paused-destination"
    POST_MIGRATION_DELETE = "post-migration-delete"


# The phases during which caching should not be happening
SHORT_CACHE_PHASES = {
    ShardStatus.MIGRATING_SYNC,
    ShardStatus.POST_MIGRATION_PAUSED_AT_SOURCE,
    ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION,
}


MIGRATION_PHASES = {
    ShardStatus.MIGRATING_COPY,
    ShardStatus.MIGRATING_SYNC,
    ShardStatus.POST_MIGRATION_PAUSED_AT_SOURCE,
}
POST_MIGRATION_PHASES = {
    ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION,
    ShardStatus.POST_MIGRATION_DELETE,
}

_caching_timeout = 0
_metadata_stores = {}
_realm_cache = {}


def _get_realm_coll():
    return get_controlling_db().realms


def _get_shards_coll():
    return get_controlling_db().shards


class ShardMetadataStore(object):
    """A store of all the shard metadata for a particular realm.
    This is NOT thread safe.

    We want to cache as many lookups as possible
    We have generic shard information
     - This should be cached for as long as possible
     - If a single shard is being moved then this part needs to be refreshed
    We also have specific queries for specific shards
     - These should be cached unless they are actively migrating
    """
    def __init__(self, realm):
        assert isinstance(realm, dict)
        self._cache = {}
        self.realm = realm
        self._global_timeout = 0
        self._in_flux = None


    def get_single_shard_metadata(self, shard_key):
        if not self._cache_entry_is_valid(shard_key):
            self._refresh_single_shard_metadata(shard_key)
        return self._cache[shard_key][0]


    def _cache_entry_is_valid(self, shard_key):
        now = time.time()
        return (
            self._in_flux != shard_key
            and shard_key in self._cache
            and self._cache[shard_key][1] > now)


    def get_all_shard_metadata(self):
        now = time.time()
        if self._global_timeout < now:
            self._refresh_all_shard_metadata()
        elif self._in_flux:
            self._refresh_single_shard_metadata(self._in_flux)
            
        return {
            shard_key: metadata for
            shard_key, (metadata, _) in self._cache.iteritems()
        }

    
    def _refresh_single_shard_metadata(self, shard_key):
        global _caching_timeout
        shards = list(self._query_shards_collection(shard_key))

        generic_expiry = time.time() + _caching_timeout
        if shards:
            shard, = shards
            if shard['status'] in SHORT_CACHE_PHASES:
                self._in_flux = shard['shard_key']
                expiry = 0
            else:
                expiry = generic_expiry

            self._cache[shard['shard_key']] = (shard, expiry)
        else:
            shard = {
                'location': self.realm['default_dest'],
                'status': ShardStatus.AT_REST,
                'realm': self.realm['name'],
            }
            self._cache[shard_key] = (shard, generic_expiry)

    
    def _refresh_all_shard_metadata(self):
        global _caching_timeout
        cursor = self._query_shards_collection()
        self._global_timeout = time.time() + _caching_timeout
        self._in_flux = None

        for shard in cursor:
            if shard['status'] in SHORT_CACHE_PHASES:
                assert not self._in_flux, "Multiple shards in motion"
                self._in_flux = shard['shard_key']
                expiry = 0
            else:
                expiry = self._global_timeout

            self._cache[shard['shard_key']] = (shard, expiry)


    def _query_shards_collection(self, shard_key=None):
        shards_coll = _get_shards_coll()
        query = {'realm': self.realm['name']}
        if shard_key:
            query['shard_key'] = shard_key
        return shards_coll.find(query)


class LocationMetadata(object):
    def __init__(self, location):
        assert location.count('/') == 1, \
            "Location must be of the form cluster/db and not %s" % location
        self.location = location
        self.contains = []
        self.excludes = []


    def __repr__(self):
        contains = ",".join(str(c) for c in self.contains[:5])
        if len(self.contains) >= 5:
            contains += "..."
        return "LocationMetadata(%s, contains: [%s], excludes: %s)" % (
            self.location, contains, self.excludes)



def _get_location_for_shard(realm, shard_key):
    """Gets the locations for the given shard. The result will be a single
    LocationMetadata object.
    """
    shard = _get_metadata_for_shard(realm, shard_key)

    status = shard['status']
    if status in POST_MIGRATION_PHASES:
        location = LocationMetadata(shard['new_location'])
        location.contains.append(shard_key)
    else:
        location = LocationMetadata(shard['location'])
        location.contains.append(shard_key)
    return location


def _get_metadata_store(realm):
    global _metadata_stores
    realm_name = realm['name']
    if realm_name not in _metadata_stores:
        _metadata_stores[realm_name] = ShardMetadataStore(realm)
    return _metadata_stores[realm_name]


def _get_shard_metadata_for_realm(realm):
    return _get_metadata_store(realm).get_all_shard_metadata()


def _get_metadata_for_shard(realm, shard_key):
    return _get_metadata_store(realm).get_single_shard_metadata(shard_key)


def _get_all_locations_for_realm(realm):
    """Gets all the locations for the given realm. The results will be of the
    form:
    
        { location: LocationMetadata(...) }

    The excludes is a list of keys that need to be excluded from any query
    performed against that location.
    """
    shards = _get_shard_metadata_for_realm(realm)
    locations = {}
    for shard in shards.itervalues():
        location = shard['location']
        if location not in locations:
            locations[location] = LocationMetadata(location)
        if 'new_location' in shard:
            new_location = shard['new_location']
            if new_location not in locations:
                locations[new_location] = LocationMetadata(new_location)

        status = shard['status']
        if status in MIGRATION_PHASES:
            locations[shard['new_location']].excludes.append(shard['shard_key'])
            locations[shard['location']].contains.append(shard['shard_key'])
        elif status in POST_MIGRATION_PHASES:
            locations[shard['location']].excludes.append(shard['shard_key'])
            locations[shard['new_location']].contains.append(shard['shard_key'])
        else:
            if 'shard_key' in shard:
                locations[shard['location']].contains.append(shard['shard_key'])

    if realm['default_dest'] not in locations:
        locations[realm['default_dest']] = LocationMetadata(
            realm['default_dest'])


    return dict(locations)


def _get_realm_for_collection(collection_name):
    global _realm_cache, _caching_timeout
    now = time.time()
    realm, expiry = _realm_cache.get(collection_name, (None, 0))
    if expiry <= now:
        realms_coll = _get_realm_coll()
        try:
            realm = realms_coll.find({'collection': collection_name})[0]
        except IndexError:
            raise Exception(
                'Realm for collection %s does not exist' % collection_name)
        expiry = now + _caching_timeout
        _realm_cache[collection_name] = realm, expiry
    return _realm_cache[collection_name][0]


def _get_realm_by_name(realm_name):
    realms_coll = _get_realm_coll()
    try:
        return realms_coll.find({'name': realm_name})[0]
    except IndexError:
        raise Exception(
            'Realm named %s does not exist' % realm_name)


def get_caching_duration():
    return _caching_timeout


def activate_caching(timeout):
    """Activates caching of metadata.

    :param int timeout: Number of seconds to cache metadata for.

    Caching is generally a good thing. However, during a migration there will be
    a pause equal to whatever the caching timeout. This is to avoid stale reads
    and writes when the source of truth for a shard changes location.
    """
    global _caching_timeout, _metadata_stores, _realm_cache
    _caching_timeout = timeout

    # Blank out the metadata stores as changing the timeout will really mess
    # up everything in them
    _metadata_stores = {}
    _realm_cache = {}


def wipe_metadata():
    """Wipes all metadata. Should only be used during testing. There is no undo.

    Wipes caches as well.
    """
    _get_realm_coll().remove()
    _get_shards_coll().remove()
    _get_cluster_coll().remove()

    _cluster_uri_cache.clear()
    _realm_cache.clear()
    _metadata_stores.clear()


def are_migrations_happening():
    """Returns True if any migrations are happening. Otherwise, False.
    """
    coll = _get_shards_coll()
    states = [
        ShardStatus.MIGRATING_COPY,
        ShardStatus.MIGRATING_SYNC,
        ShardStatus.POST_MIGRATION_PAUSED_AT_SOURCE,
        ShardStatus.POST_MIGRATION_PAUSED_AT_DESTINATION,
        ShardStatus.POST_MIGRATION_DELETE,
    ]
    return coll.find({'status': {'$in': states}}).count() > 0



from shardmonster import api, sharder
from shardmonster.tests.base import ShardingTestCase

class TestSharder(ShardingTestCase):
    def setUp(self):
        api.activate_caching(0.5)
        super(TestSharder, self).setUp()


    def tearDown(self):
        # Deactivate caching by setting a 0 timeout
        api.activate_caching(0)
        super(TestSharder, self).tearDown()


    def test_basic_copy(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        doc1 = {'x': 1, 'y': 1}
        doc1['_id'] = self.db1.dummy.insert(doc1)

        api.start_migration('dummy', 1, "dest2/test_sharding")

        sharder._do_copy('dummy', 1)

        # The data should now be on the second database
        doc2, = self.db2.dummy.find({})
        self.assertEquals(doc1, doc2)


    def test_sync_after_copy(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.start_migration('dummy', 1, "dest2/test_sharding")

        # Mimic the state the shard would be in after a document was copied
        # from one location to another
        doc1 = {'x': 1, 'y': 1}
        doc1['_id'] = self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc1)

        # Get the initial oplog position, do an update and then sync from the
        # initial position
        initial_oplog_pos = sharder._get_oplog_pos('dummy', 1)
        self.db1.dummy.update({'x': 1}, {'$inc': {'y': 1}})
        api.set_shard_to_migration_status(
            'dummy', 1, api.ShardStatus.MIGRATING_SYNC)
        sharder._sync_from_oplog('dummy', 1, initial_oplog_pos)

        # The data on the second database should now reflect the update that
        # went through
        doc2, = self.db2.dummy.find({})
        self.assertEquals(2, doc2['y'])


    def test_delete_after_migration(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.start_migration('dummy', 1, "dest2/test_sharding")

        # Mimic the state the shard would be in after a document was copied
        # from one location to another
        doc1 = {'x': 1, 'y': 1}
        doc1['_id'] = self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc1)

        api.set_shard_to_migration_status(
            'dummy', 1, api.ShardStatus.POST_MIGRATION_DELETE)
        sharder._delete_source_data('dummy', 1)

        # The data on the first database should now be gone and the data
        # on the second database should be ok.
        self.assertEquals(0, self.db1.dummy.find({}).count())
        doc1_actual, = self.db2.dummy.find({})
        self.assertEquals(doc1, doc1_actual)


    def test_sync_ignores_other_collection(self):
        api.set_shard_at_rest('dummy', 1, "dest1/test_sharding")
        api.start_migration('dummy', 1, "dest2/test_sharding")

        # Mimic the state the shard would be in after a document was copied
        # from one location to another
        doc1 = {'x': 1, 'y': 1}
        doc1['_id'] = self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc1)

        # Get the initial oplog position, do an update to a different collection
        # and then sync from the initial position
        initial_oplog_pos = sharder._get_oplog_pos('dummy', 1)
        self.db1.other_coll.insert(doc1)
        self.db1.other_coll.update({'x': 1}, {'$inc': {'y': 1}})
        api.set_shard_to_migration_status(
            'dummy', 1, api.ShardStatus.MIGRATING_SYNC)
        sharder._sync_from_oplog('dummy', 1, initial_oplog_pos)

        # The data on the second database should be in the same state as it
        # was before
        doc2, = self.db2.dummy.find({})
        self.assertEquals(1, doc2['y'])


    def test_sync_uses_correct_connection(self):
        """This tests for a bug found during a rollout. The connection for the
        metadata was assumed to be the same connection as the source data was
        going to be coming from. This is *not* always the case.
        """
        # To test this a migration from new to old will expose the bug
        api.set_shard_at_rest('dummy', 1, "dest2/test_sharding")
        api.start_migration('dummy', 1, "dest1/test_sharding")

        # Mimic the state the shard would be in after a document was copied
        # from one location to another
        doc1 = {'x': 1, 'y': 1}
        doc1['_id'] = self.db1.dummy.insert(doc1)
        self.db2.dummy.insert(doc1)

        # Get the initial oplog position, do an update and then sync from the
        # initial position
        initial_oplog_pos = sharder._get_oplog_pos('dummy', 1)
        self.db2.dummy.update({'x': 1}, {'$inc': {'y': 1}})
        api.set_shard_to_migration_status(
            'dummy', 1, api.ShardStatus.MIGRATING_SYNC)
        sharder._sync_from_oplog('dummy', 1, initial_oplog_pos)

        # The data on the first database should now reflect the update that
        # went through
        doc2, = self.db1.dummy.find({})
        self.assertEquals(2, doc2['y'])



import pymongo
import unittest

import test_settings
from shardmonster import api, connection as connection_module, metadata


def _is_same_mongo(conn1, conn2):
    uri_info1 = pymongo.uri_parser.parse_uri(conn1['uri'])
    uri_info2 = pymongo.uri_parser.parse_uri(conn2['uri'])
    return set(uri_info1['nodelist']) == set(uri_info2['nodelist'])


class ShardingTestCase(unittest.TestCase):
    def setUp(self):
        self._prepare_connections()
        self._clean_data_before_tests()
        self._prepare_clusters()
        self._prepare_realms()
        api.create_indices()


    def tearDown(self):
        self.conn1.close()
        self.conn2.close()


    def _create_connection(self, conn_settings):
        conn = connection_module._connect_to_mongo(conn_settings['uri'])
        db_name = conn_settings['db_name']
        db = conn[db_name]
        return conn, db


    def _prepare_connections(self):
        if _is_same_mongo(test_settings.CONN1, test_settings.CONN2):
            raise Exception('Must use two different Mongo servers for testing!')
        self.conn1, self.db1 = self._create_connection(
            test_settings.CONN1)
        self.conn2, self.db2 = self._create_connection(
            test_settings.CONN2)
        api.connect_to_controller(
            test_settings.CONTROLLER['uri'],
            test_settings.CONTROLLER['db_name'],
        )

        # Wipe the connections that are in the cache in shardmonster
        for connection in connection_module._connection_cache.values():
            connection.close()
        connection_module._connection_cache = {}
        connection_module._cluster_uri_cache = {}
        api._collection_cache = {}
        metadata._metadata_stores = {}


    def _clean_data_before_tests(self):
        self.db1.client.drop_database(self.db1.name)
        self.db2.client.drop_database(self.db2.name)
        api._reset_sharding_info()


    def _prepare_clusters(self):
        api.add_cluster('dest1', test_settings.CONN1['uri'])
        api.add_cluster('dest2', test_settings.CONN2['uri'])


    def _prepare_realms(self):
        api.create_realm(
            'dummy', 'x', 'dummy',
            'dest1/%s' % test_settings.CONN1['db_name'])



from shardmonster.connection import (
    add_cluster, connect_to_controller, _get_cluster_coll, get_cluster_uri,
    parse_location)
from shardmonster.metadata import (
    _get_location_for_shard, _get_realm_coll, _get_realm_by_name,
    _get_realm_for_collection, _get_shards_coll, ShardStatus, activate_caching,
    get_caching_duration)
from shardmonster import operations

__all__ = [
    "activate_caching", "connect_to_controller", "get_caching_duration",
    "add_cluster", "set_shard_at_rest"]

_collection_cache = {}


def create_indices():
    realm_coll = _get_realm_coll()
    realm_coll.ensure_index([('name', 1)], unique=True)
    realm_coll.ensure_index([('collection', 1)], unique=True)

    shards_coll = _get_shards_coll()
    shards_coll.ensure_index(
        [('realm', 1), ('shard_key', 1)], unique=True)
    shards_coll.ensure_index([('status', 1)])

    cluster_coll = _get_cluster_coll()
    cluster_coll.ensure_index([('name', 1)], unique=True)


def create_realm(realm, shard_field, collection_name, default_dest):
    _get_realm_coll().insert({
        'name': realm,
        'shard_field': shard_field,
        'collection': collection_name,
        'default_dest': default_dest})


def ensure_realm_exists(name, shard_field, collection_name, default_dest):
    """Ensures that a realm of the given name exists and matches the expected
    settings.

    :param str name: The name of the realm
    :param shard_field: The field in documents that should be used as the shard
        field. The only supported values that can go in this field are strings
        and integers.
    :param str collection_name: The name of the collection that this realm
        corresponds to. In general, the collection name should match the realm
        name.
    :param str default_dest: The default destination for any data that isn't
        explicitly sharded to a specific location.
    :return: None
    """
    coll = _get_realm_coll()

    cursor = coll.find({'name': name})
    if cursor.count():
        # realm with this name already exists
        existing = cursor[0]
        if (existing['shard_field'] != shard_field
            or existing['collection'] != collection_name
            or existing['default_dest'] != default_dest):
            raise Exception('Cannot change realm')
        else:
            return
        
    cursor = coll.find({'collection': collection_name})
    if cursor.count():
        # realm for this collection already exists
        existing = cursor[0]
        if (existing['shard_field'] != shard_field
            or existing['name'] != name
            or existing['default_dest'] != default_dest):
            raise Exception(
                'Realm for collection %s already exists' % collection_name)
        else:
            return

    create_realm(name, shard_field, collection_name, default_dest)


def _assert_valid_location(location):
    cluster_name, _ = parse_location(location)
    # Attempting to get the URI for a non-existant cluster will throw an
    # exception
    get_cluster_uri(cluster_name)
        

def set_shard_at_rest(realm, shard_key, location, force=False):
    """Marks a shard as being at rest in the given location. This is used for
    initiating shards in preparation for migration. Unless force is True this
    will raise an exception if a shard is already at rest in a specific
    location.

    :param str realm: The name of the realm for the shard
    :param shard_key: The key of the shard
    :param str location: The location that the data is at (or should be in the
        case of a brand new shard)
    :param bool force: Force a shard to be placed at rest in a specific location
        even if it has already been placed somewhere.
    :return: None
    """
    _assert_valid_location(location)
    
    shards_coll = _get_shards_coll()

    query = {'realm': realm, 'shard_key': shard_key}
    if shards_coll.find(query).count() and not force:
        raise Exception(
            'Shard with key %s has already been placed. Use force=true if '
            'you really want to do this' % shard_key)

    shards_coll.update(query,
        {
            '$set': {
                'location': location,
                'status': ShardStatus.AT_REST,
            },
            '$unset': {
                'new_location': 1,
            },
        },
        upsert=True)


def set_shard_to_migration_status(realm, shard_key, status):
    """Marks a shard as being at a specific migration status.
    """
    shards_coll = _get_shards_coll()
    shards_coll.update(
        {'realm': realm, 'shard_key': shard_key},
        {'$set': {'status': status}}
    )


def start_migration(realm_name, shard_key, new_location):
    """Marks a shard as being in the process of being migrated.
    """
    shards_coll = _get_shards_coll()
    realm = _get_realm_by_name(realm_name)
    existing_location = _get_location_for_shard(realm, shard_key)
    if existing_location.location == new_location:
        raise Exception('Shard is already at %s' % new_location)

    shards_coll.update(
        {'realm': realm_name, 'shard_key': shard_key},
        {'$set': {
            'status': ShardStatus.MIGRATING_COPY,
            'new_location': new_location,
        }},
    )


def _reset_sharding_info():
    """Wipes all shard info. For internal test use only.
    """
    _get_cluster_coll().remove({})
    _get_realm_coll().remove({})
    _get_shards_coll().remove({})


class ShardAwareCollectionProxy(object):
    def __init__(self, collection_name):
        self.collection_name = collection_name
        self._with_options = {}

    def find(self, *args, **kwargs):
        return operations.multishard_find(
            self.collection_name,
            with_options=self._with_options, *args, **kwargs)

    def find_one(self, *args, **kwargs):
        return operations.multishard_find_one(
            self.collection_name,
            with_options=self._with_options, *args, **kwargs)

    def update(self, *args, **kwargs):
        return operations.multishard_update(
            self.collection_name,
            with_options=self._with_options, *args, **kwargs)

    def insert(self, *args, **kwargs):
        return operations.multishard_insert(
            self.collection_name,
            with_options=self._with_options, *args, **kwargs)

    def remove(self, *args, **kwargs):
        return operations.multishard_remove(
            self.collection_name,
            with_options=self._with_options, *args, **kwargs)

    def save(self, *args, **kwargs):
        return operations.multishard_save(
            self.collection_name,
            with_options=self._with_options, *args, **kwargs)

    def with_options(self, **kwargs):
        new_collection = ShardAwareCollectionProxy(self.collection_name)
        new_collection._with_options = self._with_options.copy()
        new_collection._with_options.update(**kwargs)
        return new_collection

    def ensure_index(self, *args, **kwargs):
        return operations.multishard_ensure_index(
            self.collection_name, *args, **kwargs)

    def aggregate(self, *args, **kwargs):
        return operations.multishard_aggregate(
            self.collection_name, *args, **kwargs)


def make_collection_shard_aware(collection_name):
    """Returns a new object that proxies the given collection and makes it
    shard aware.
    """
    return ShardAwareCollectionProxy(collection_name)


def where_is(collection_name, shard_key):
    """Returns a string of the form cluster/database that says where a
    particular shard of data resides.

    :param collection_name: The collection name for the shard
    :param shard_key: The shard key to look for
    """
    realm = _get_realm_for_collection(collection_name)
    location = _get_location_for_shard(realm, shard_key)
    return location.location



from shardmonster.api import (
    activate_caching, connect_to_controller, ensure_realm_exists,
    make_collection_shard_aware, set_shard_at_rest, where_is)
from shardmonster.connection import ensure_cluster_exists
from shardmonster.metadata import wipe_metadata
from shardmonster.sharder import do_migration

__all__ = [
    'activate_caching', 'connect_to_controller', 'do_migration',
    'ensure_cluster_exists', 'ensure_realm_exists',
    'make_collection_shard_aware', 'set_shard_at_rest',
    'where_is', 'wipe_metadata', 'VERSION',
]

VERSION = (0, 4, 2)
